#include "netwin.h"
#include "preferences.h"
#include "codenet.h"

NetWin::NetWin(QWidget *parent)
: QDialog(parent,Qt::Tool),
  m_codeNet(0),
  m_netDrive(0)
{
  setWindowTitle("DiskImagery64: " + tr("Network"));
  QGridLayout *layout = new QGridLayout(this);
  
  // log view
  m_logView = new QTextEdit(this);
  m_logView->setReadOnly(true);
  m_logView->setLineWrapMode(QTextEdit::NoWrap);
  QFont smallFont("Monaco");
  smallFont.setPointSize(9);
  m_logView->setFont(smallFont);
  
  // packet counter
  m_packetsInLabel = new QLabel(this);
  m_packetsOutLabel = new QLabel(this);
  
  m_stopButton = new QToolButton(this);
  m_stopButton->setText(tr("Abort"));
  connect(m_stopButton,SIGNAL(clicked()),this,SLOT(stopNet()));
  
  layout->addWidget(m_logView,0,0,1,3);
  layout->addWidget(m_packetsInLabel,1,0);
  layout->addWidget(m_packetsOutLabel,1,1);
  layout->addWidget(m_stopButton,1,2);
  layout->setMargin(2);
  
  resetCounters();
  loadSettings();
}

NetWin::~NetWin()
{
  saveSettings();
  stopNet();
}

// ----- Settings -----

void NetWin::saveSettings()
{
  QSettings settings;
  settings.beginGroup("NetWin");
  
  // geometry
  settings.setValue("geometry",geometry());
  // visible
  settings.setValue("visible",isVisible());
  
  settings.endGroup();
}

void NetWin::loadSettings()
{
  QSettings settings;
  settings.beginGroup("NetWin");

  // geometry
  QRect rect = settings.value("geometry",QRect(400,100,200,100)).toRect();
  move(rect.topLeft());
  resize(rect.size());
  // visible
  bool visible = settings.value("visible",false).toBool();
  if(visible)
    show();
  else
    hide();
  
  settings.endGroup();
}

// ----- UI -----

void NetWin::logMessage(const QString &msg)
{
  QDateTime now = QDateTime::currentDateTime();
  QLocale locale;
  QString header = locale.toString(now.date(),QLocale::ShortFormat) + " " +
                   locale.toString(now.time(),QLocale::ShortFormat) + " ";
  m_logView->append(header + msg);
}

void NetWin::logService(NetService *service)
{
  connect(service,SIGNAL(networkEvent(const QString &)),
          this,SLOT(logMessage(const QString &)));
  connect(service,SIGNAL(sentPacket(int)),
          this,SLOT(reportOutPacket(int)));
  connect(service,SIGNAL(receivedPacket(int)),
          this,SLOT(reportInPacket(int)));
}

void NetWin::resetCounters()
{
  m_packetsIn = 0;
  m_packetsInSize = 0;
  m_packetsOut = 0;
  m_packetsOutSize = 0;
  updateCounters();
}

void NetWin::updateCounters()
{
  QString inMsg;
  inMsg.sprintf("In: %d/%d bytes",m_packetsIn,m_packetsInSize);
  m_packetsInLabel->setText(inMsg);
  QString outMsg;
  outMsg.sprintf("Out: %d/%d bytes",m_packetsOut,m_packetsOutSize);
  m_packetsOutLabel->setText(outMsg);
}

void NetWin::reportInPacket(int size)
{
  m_packetsIn++;
  m_packetsInSize += size;
  updateCounters();
}

void NetWin::reportOutPacket(int size)
{
  m_packetsOut++;
  m_packetsOutSize += size;
  updateCounters();
}

// ----- Commands -----

void NetWin::stopNet()
{
  if(m_codeNet!=0) {
    delete m_codeNet;
    m_codeNet = 0;
  }
  if(m_netDrive!=0) {
    delete m_netDrive;
    m_netDrive = 0;
  }
  resetCounters();
}

void NetWin::runProgram(quint16 addr,const QByteArray &data)
{
  startCodeNet();
  
  m_codeNet->sendData(addr,data);
  if(addr==0x0801)
    m_codeNet->execRun();
  else
    m_codeNet->execJump(addr);
}

void NetWin::shareFiles(const CBMFileList &files)
{
  startNetDrive();
  
  m_netDrive->shareFiles(files);
}

void NetWin::startCodeNet()
{
  if(m_codeNet!=0)
    return;
  
  AddrPair addrPair;
  Preferences::getNetworkDefaults(addrPair);
  m_codeNet = new CodeNetService(addrPair);
  logService(m_codeNet);
}

void NetWin::startNetDrive()
{
  if(m_netDrive!=0)
    return;

  AddrPair addrPair;
  Preferences::getNetworkDefaults(addrPair);
  m_netDrive = new NetDriveService(addrPair);
  logService(m_netDrive);   
}
