#include "netservice.h"

NetService::NetService(const AddrPair &addrPair,const QString &name)
: m_addrPair(addrPair),
  m_name(name),
  m_doStop(false),
  m_failedFatal(false)
{
}

NetService::~NetService()
{
  stop();
}

void NetService::addTask(NetTask *task,bool cancelAll)
{
  bool failed;
  {
    QMutexLocker locker(&m_mutex);
    
    // clean all existing tasks first
    if(cancelAll) {
      for(int i=0;i<m_tasks.size();i++) {
        m_tasks.at(i)->doStop();
      }
    }
    
    // add new task
    m_tasks.enqueue(task);
    
    // get failed status
    failed = m_failedFatal;
  }
  
  // do not try to run if already failed
  if(failed)
    return;
  
  // start thread if not running
  if(!isRunning())
    start();
}

void NetService::stop()
{
  for(int i=0;i<10;i++) {
    // if not running then quit
    if(!isRunning())
      return; 
    // set stop flag
    {
      QMutexLocker locker(&m_mutex);
      m_doStop = true;
    }
    // wait for thread to stop
    wait(1000);
  }

  // last chance: terminate it
  terminate();
}

void NetService::reset()
{
  QMutexLocker locker(&m_mutex);
  m_failedFatal = false;
}

void NetService::run()
{
  // service already failed
  {
    QMutexLocker locker(&m_mutex);
    if(m_failedFatal)
      return;
  }
  
  // create host
  NetHost *host = new NetHost;
  connect(host,SIGNAL(sentPacket(int)),this,SIGNAL(sentPacket(int)));
  connect(host,SIGNAL(receivedPacket(int)),this,SIGNAL(receivedPacket(int)));
  
  // try to bind host  
  if(!host->bind(m_addrPair)) {
    emit networkEvent(tr("ERROR: ")+host->errorString());
    {
      QMutexLocker locker(&m_mutex);
      m_failedFatal = true;
    }
    delete host;
    return;
  }
  emit networkEvent(tr("Bound service '%1' to %2:%3")
                    .arg(m_name)
                    .arg(m_addrPair.myAddr.toString())
                    .arg(QString::number(m_addrPair.myPort)));
  emit networkEvent(tr("  C64 Address: %1:%2")
                    .arg(m_addrPair.c64Addr.toString())
                    .arg(QString::number(m_addrPair.c64Port)));
  
  // main loop
  int idleCount = 0;
  int retryCount = 0;
  forever {
    NetTask *currentTask;
    {
      // enter critical section
      QMutexLocker locker(&m_mutex);

      // stop service?
      if(m_doStop) {
        m_doStop = false;
        break;
      }
      
      // remove stopped tasks
      while(!m_tasks.empty() && m_tasks.head()->isStopped()) {
        NetTask *oldTask = m_tasks.dequeue();
        delete oldTask;
      }
      
      // get current task
      if(m_tasks.empty())
        currentTask = 0;
      else {
        currentTask = m_tasks.head();
      }
    }

    if(host->isBound()) {
      if(currentTask!=0) {
        idleCount = 0;
      
        // perform task
        QString resultEvent;
        NetTask::Result result;
        result = currentTask->run(*host,resultEvent);
        if(resultEvent!="")
          emit networkEvent(resultEvent);

        // retry command?
        if(result==NetTask::RETRY) {
          if(retryCount==5) {
            retryCount = 0;
            result = NetTask::ERROR;
            emit networkEvent(tr("Retries failed... time out?"));
          } else {
            retryCount++;
            emit networkEvent(tr("Retrying task (attempt #%1)").arg(QString::number(retryCount)));
            snooze(host);
          }
        }
        // an error flushes the tasks
        if(result==NetTask::ERROR) {
          flushTasks();
          //flushPackets(host);
          snooze(host);
        }
        // ready removes the current entry and goes on with the next one
        else if(result==NetTask::READY) {
          currentTask->doStop();
        }
      } else {
        // too long -> close service
        if(idleCount==100) {
          emit networkEvent(tr("Nothing to do..."));
          break;
        }
      
        // no task currently available... we are idle
        idleCount++;
        snooze(host);
      }
    } else {
      host->close();
      host->bind(m_addrPair);
      emit networkEvent(tr("Waiting for binding..."));
      msleep(2000);
    }
  }
  
  flushTasks();
  
  // close host
  host->close();
  emit networkEvent(tr("Closed service '%1'").arg(m_name));
  delete host;
}

void NetService::snooze(NetHost *host)
{
  // either sleep  
  if(host->hasPackets()) {
    emit networkEvent(tr("Packets pending..."));
    msleep(500);
  } else {
    host->waitForPacket(500);
  }
}

void NetService::flushTasks()
{
  // clean up task list if service was stopped
  int unfinishedTasks = 0;
  {
    QMutexLocker locker(&m_mutex);
    while(!m_tasks.empty()) {
      NetTask *task = m_tasks.dequeue();
      delete task;
      unfinishedTasks++;
    }
  }
  if(unfinishedTasks>0)
    emit networkEvent(tr("Flushed %1 tasks")
                      .arg(QString::number(unfinishedTasks)));
}

void NetService::flushPackets(NetHost *host)
{
  // flush packets
  int num = host->flushPackets(1000);
  if(num>0)
    emit networkEvent(tr("Flushed %1 packets")
                      .arg(QString::number(num)));
}
